import re
from dataclasses import dataclass
from datetime import UTC, datetime
from uuid import UUID

from sqlalchemy import DateTime, event
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import Mapped
from sqlalchemy.testing.schema import mapped_column
from sqlalchemy_utils import UUIDType
from uuid6 import uuid7

from fastgear.common.database.sqlalchemy.base import Base


@dataclass
class BaseEntity(Base):
    __abstract__ = True

    id: Mapped[UUID] = mapped_column(UUIDType(), primary_key=True, default=uuid7)
    created_at: Mapped[datetime] = mapped_column(DateTime(timezone=True), nullable=False)
    updated_at: Mapped[datetime] = mapped_column(DateTime(timezone=True), nullable=False)

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

    @declared_attr
    def __tablename__(self) -> str:
        """Generates the name of the table that will be created in the database for a given BaseEntity class.

        The table name is derived from the class name by converting it from CamelCase to snake_case.

        Returns:
            str: The generated table name in snake_case.

        """
        return "_".join(re.findall("[A-Z][^A-Z]*", self.__name__)).lower()


@event.listens_for(BaseEntity, "before_insert", propagate=True)
def set_before_insert(mapper, connection, target: BaseEntity) -> None:
    """Event listener that sets the `created_at` and `updated_at` timestamps before inserting a new record.

    This function is triggered before a new `BaseEntity` record is inserted into the database.

    It ensures that the `created_at` and `updated_at` fields are set to the current datetime if they are not already
    set.

    Args:
        mapper: The SQLAlchemy mapper.
        connection: The database connection.
        target (BaseEntity): The instance of `BaseEntity` being inserted.

    """
    if not target.created_at:
        target.created_at = datetime.now(UTC)
    if not target.updated_at or target.updated_at < target.created_at:
        target.updated_at = target.created_at


@event.listens_for(BaseEntity, "before_update", propagate=True)
def set_before_update(mapper, connection, target: BaseEntity) -> None:
    target.updated_at = datetime.now(UTC)



from enum import StrEnum
from pathlib import Path

from pydantic_settings import BaseSettings, PydanticBaseSettingsSource, TomlConfigSettingsSource


class TomlBaseSettings(BaseSettings):
    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        return env_settings, TomlConfigSettingsSource(settings_cls)

    @staticmethod
    def get_toml_files(config_dir: Path, env_enum: type[StrEnum]) -> list[str]:
        candidates: list[Path] = [
            config_dir / "env.toml",
            config_dir / "env.local.toml",
        ]

        for env in env_enum:
            candidates.append(config_dir / f"env.{env.value}.toml")
            candidates.append(config_dir / f"env.{env.value}.local.toml")

        seen: set[Path] = set()
        files: list[str] = []
        for p in candidates:
            if not p.exists():
                continue
            if p in seen:
                continue
            seen.add(p)
            files.append(str(p))

        return files



from starlette.status import (
    HTTP_400_BAD_REQUEST,
    HTTP_401_UNAUTHORIZED,
    HTTP_403_FORBIDDEN,
    HTTP_404_NOT_FOUND,
    HTTP_422_UNPROCESSABLE_CONTENT,
    HTTP_429_TOO_MANY_REQUESTS,
)

from fastgear.types.custom_base_exception import CustomBaseException


class BadRequestException(CustomBaseException):
    def __init__(self, msg: str, loc: list[str] = None, _type: str = "Bad Request") -> None:
        if loc is None:
            loc = []
        self.status_code = HTTP_400_BAD_REQUEST
        super().__init__(msg, loc, _type)


class UnauthorizedException(CustomBaseException):
    def __init__(self, msg: str, loc: list[str] = None, _type: str = "unauthorized") -> None:
        if loc is None:
            loc = []
        self.status_code = HTTP_401_UNAUTHORIZED
        super().__init__(msg, loc, _type)


class ForbiddenException(CustomBaseException):
    def __init__(self, msg: str, loc: list[str] = None, _type: str = "forbidden") -> None:
        if loc is None:
            loc = []
        self.status_code = HTTP_403_FORBIDDEN
        super().__init__(msg, loc, _type)


class NotFoundException(CustomBaseException):
    def __init__(self, msg: str, loc: list[str] = None, _type: str = "Not Found") -> None:
        if loc is None:
            loc = []
        self.status_code = HTTP_404_NOT_FOUND
        super().__init__(msg, loc, _type)


class UnprocessableEntityException(CustomBaseException):
    def __init__(
        self, msg: str, loc: list[str] = None, _type: str = "Unprocessable Entity"
    ) -> None:
        if loc is None:
            loc = []
        self.status_code = HTTP_422_UNPROCESSABLE_CONTENT
        super().__init__(msg, loc, _type)


class DuplicateValueException(CustomBaseException):
    def __init__(self, msg: str, loc: list[str] = None, _type: str = "Duplicate Value") -> None:
        if loc is None:
            loc = []
        self.status_code = HTTP_422_UNPROCESSABLE_CONTENT
        super().__init__(msg, loc, _type)


class RateLimitException(CustomBaseException):
    def __init__(self, msg: str, loc: list[str] = None, _type: str = "Rate Limit") -> None:
        if loc is None:
            loc = []
        self.status_code = HTTP_429_TOO_MANY_REQUESTS
        super().__init__(msg, loc, _type)


CustomHTTPExceptionType = (
    BadRequestException
    | UnauthorizedException
    | ForbiddenException
    | NotFoundException
    | UnprocessableEntityException
    | DuplicateValueException
    | RateLimitException
)



from abc import ABC, abstractmethod
from contextvars import ContextVar

from sqlalchemy import create_engine
from sqlalchemy.ext.asyncio import (
    AsyncSession,
    AsyncSessionTransaction,
    async_sessionmaker,
    create_async_engine,
)
from sqlalchemy.orm import Session, SessionTransaction, sessionmaker

SyncSessionType = Session | SessionTransaction
AsyncSessionType = AsyncSession | AsyncSessionTransaction
AllSessionType = SyncSessionType | AsyncSessionType

# Unified context variable for both sync and async sessions
db_session: ContextVar[AllSessionType | None] = ContextVar("db_session", default=None)


class AbstractDatabaseSessionFactory(ABC):
    def __init__(self) -> None:
        """Initializes the DatabaseSessionFactory with the given database URL."""

    @abstractmethod
    def get_session(self) -> Session | AsyncSession:
        """Creates a new database session.

        Returns:
            SyncSessionType: A new SQLAlchemy session.
        """


class SyncDatabaseSessionFactory(AbstractDatabaseSessionFactory):
    def __init__(self, database_url: str) -> None:
        """Initializes the DatabaseSessionFactory with the given database URL.

        Args:
            database_url (str): The URL of the database to connect to.
        """
        super().__init__()
        self.engine = create_engine(database_url)
        self.SessionLocal = sessionmaker(
            self.engine, autocommit=False, autoflush=False, expire_on_commit=False
        )

    def get_session(self) -> Session:
        """Creates a new database session.

        Returns:
            Session: A new SQLAlchemy session.

        """
        return self.SessionLocal()


class AsyncDatabaseSessionFactory(AbstractDatabaseSessionFactory):
    def __init__(self, database_url: str) -> None:
        """Initializes the AsyncDatabaseSessionFactory with the given database URL.

        Args:
            database_url (str): The URL of the database to connect to.
        """
        super().__init__()
        self.async_engine = create_async_engine(database_url)
        self.AsyncSessionLocal = async_sessionmaker(
            self.async_engine,
            autocommit=False,
            autoflush=False,
            expire_on_commit=False,
            class_=AsyncSession,
        )

    def get_session(self) -> AsyncSession:
        """Creates a new asynchronous database session.

        Returns:
            AsyncSession: A new SQLAlchemy asynchronous session.
        """
        return self.AsyncSessionLocal()

    async def close_engine(self) -> None:
        """Closes the asynchronous database engine.

        This method should be called to properly dispose of the engine and release any resources.
        """
        await self.async_engine.dispose()



from .controller_decorator import controller
from .db_session_decorator import DBSessionDecorator
from .pagination_with_search_decorator import PaginationWithSearchOptions
from .simple_pagination_decorator import SimplePaginationOptions

__all__ = [
    "DBSessionDecorator",
    "PaginationWithSearchOptions",
    "SimplePaginationOptions",
    "controller",
]



from datetime import datetime

from pydantic import BaseModel, ConfigDict


class CustomBaseModel(BaseModel):
    """A custom base model that extends Pydantic's BaseModel.

    This class provides common configurations and behaviors for other data models in the application.
    """

    # model_config (ConfigDict): Pydantic v2 configuration that registers JSON encoders.
    # Maps datetime objects to ISO 8601 strings using datetime.isoformat() so datetime fields
    # are JSON-serializable (e.g. when calling .model_dump_json() or similar serializers).
    model_config = ConfigDict(json_encoders={datetime: lambda v: v.isoformat()})



from typing import Any

from sqlalchemy import (
    BinaryExpression,
    Delete,
    Select,
    String,
    Update,
    asc,
    bindparam,
    delete,
    desc,
    inspect,
    or_,
    select,
    update,
)
from sqlalchemy.orm import load_only, selectinload
from sqlalchemy_utils import cast_if

from fastgear.types.delete_options import DeleteOptions
from fastgear.types.find_many_options import FindManyOptions
from fastgear.types.find_one_options import FindOneOptions
from fastgear.types.generic_types_var import EntityType
from fastgear.types.pagination import Pagination, PaginationSearch
from fastgear.types.update_options import UpdateOptions


class StatementConstructor:
    def __init__(self, entity: EntityType) -> None:
        self.entity = entity

    def build_select_statement(
        self,
        criteria: str | FindOneOptions | FindManyOptions | Pagination = None,
        new_entity: EntityType = None,
    ) -> Select:
        """Constructs and returns a SQLAlchemy Select statement based on the provided criteria and entity.

        Args:
            criteria (str | FindOneOptions | FindManyOptions, Pagination, optional): The filter criteria to build the select
                statement. It can be a string, an instance of FindOneOptions, an instance of FindManyOptions or Pagination.
                Defaults to None.
            new_entity (EntityType, optional): A new entity type to use for the select statement.
                If not provided, the existing entity type will be used. Defaults to None.

        Returns:
            Select: The constructed SQLAlchemy Select statement.

        """
        entity = new_entity if new_entity is not None else self.entity

        if isinstance(criteria, str):
            criteria = self.build_where_from_id(criteria, entity)

        statement = select(entity)

        return self._apply_select_options(statement, entity, criteria)

    def _apply_select_options(
        self,
        statement: Select,
        entity: EntityType,
        options_dict: FindOneOptions | FindManyOptions = None,
    ) -> Select:
        """Applies various options to the given SQLAlchemy Select statement based on the provided option's dictionary.

        Args:
            statement (Select): The initial SQLAlchemy Select statement to which options will be applied.
            entity (EntityType): The entity type associated with the select statement.
            options_dict (FindOneOptions | FindManyOptions, optional): A dictionary containing various options to be
                applied to the select statement. Defaults to None.

        Returns:
            Select: The modified SQLAlchemy Select statement with the applied options.

        """
        if not options_dict:
            return statement

        options_dict = self._fix_options_dict(options_dict)

        for key, value in options_dict.items():
            match key:
                case "select":
                    statement = statement.options(load_only(*value, raiseload=True))
                case "where":
                    statement = statement.where(*value)
                case "order_by":
                    statement = statement.order_by(*value)
                case "skip":
                    statement = statement.offset(value)
                case "take":
                    statement = statement.limit(value)
                case "relations":
                    statement = statement.options(
                        *[selectinload(getattr(entity, relation)) for relation in value]
                    )
                case "with_deleted":
                    statement = statement.execution_options(with_deleted=value)
                case _:
                    raise KeyError(f"Unknown option: {key} in FindOptions")

        return statement

    def build_update_statement(
        self,
        criteria: str | UpdateOptions = None,
        *,
        new_entity: EntityType = None,
        payload: dict[str, Any],
    ) -> Update:
        entity = new_entity if new_entity is not None else self.entity

        if isinstance(criteria, str):
            criteria = self.build_where_from_id(criteria, entity)

        diff_conditions = [
            getattr(entity, k).is_distinct_from(bindparam(f"cmp_{k}")) for k in payload
        ]

        criteria["where"].append(or_(*diff_conditions))  # Only update if values are different

        statement = update(entity)

        statement = self._apply_update_options(statement, criteria)
        return statement.values(**payload)

    def _apply_update_options(
        self,
        statement: Update,
        options_dict: UpdateOptions = None,
    ) -> Update:
        """Applies various options to the given SQLAlchemy Select statement based on the provided option's dictionary.

        Args:
            statement (Update): The initial SQLAlchemy Select statement to which options will be applied.
            options_dict (UpdateOptions, optional): A dictionary containing various options to be
                applied to the update statement. Defaults to None.

        Returns:
            Update: The modified SQLAlchemy Select statement with the applied options.
        """
        if not options_dict:
            return statement

        options_dict = self._fix_options_dict(options_dict)

        for key, value in options_dict.items():
            match key:
                case "where":
                    statement = statement.where(*value)
                case _:
                    raise KeyError(f"Unknown option: {key} in UpdateOptions")

        return statement

    def build_delete_statement(
        self,
        criteria: str | DeleteOptions = None,
        *,
        new_entity: EntityType = None,
    ) -> Update:
        entity = new_entity if new_entity is not None else self.entity

        if isinstance(criteria, str):
            criteria = self.build_where_from_id(criteria, entity)

        statement = delete(entity)

        return self._apply_delete_options(statement, criteria)

    def _apply_delete_options(
        self,
        statement: Delete,
        options_dict: DeleteOptions = None,
    ) -> Delete:
        """Applies various options to the given SQLAlchemy Select statement based on the provided option's dictionary.

        Args:
            statement (Delete): The initial SQLAlchemy Select statement to which options will be applied.
            options_dict (DeleteOptions, optional): A dictionary containing various options to be
                applied to the delete statement. Defaults to None.

        Returns:
            Delete: The modified SQLAlchemy Select statement with the applied options.
        """
        if not options_dict:
            return statement

        options_dict = self._fix_options_dict(options_dict)

        for key, value in options_dict.items():
            match key:
                case "where":
                    statement = statement.where(*value)
                case _:
                    raise KeyError(f"Unknown option: {key} in UpdateOptions")

        return statement

    @staticmethod
    def extract_from_mapping(field_mapping: dict, fields: list) -> list:
        """Extracts and returns a list of items from the field mapping based on the provided fields.

        Args:
            field_mapping (dict): A dictionary mapping fields to their corresponding items.
            fields (list): A list of fields to extract items for.

        Returns:
            list: A list of items extracted from the field mapping based on the provided fields.

        """
        return [
            item
            for field in fields
            for item in (
                field_mapping.get(field, [field])
                if isinstance(field_mapping.get(field, field), list)
                else [field_mapping.get(field, field)]
            )
        ]

    @staticmethod
    def _fix_options_dict(
        options_dict: FindOneOptions | FindManyOptions | UpdateOptions,
    ) -> FindOneOptions | FindManyOptions:
        """Ensures that specific attributes in the options dictionary are lists.

        Args:
            options_dict (FindOneOptions | FindManyOptions | UpdateOptions): The options dictionary to be fixed.

        Returns:
            FindOneOptions | FindManyOptions: The fixed options dictionary with specific attributes as lists.

        """
        for attribute in ["where", "order_by", "options"]:
            if attribute in options_dict and not isinstance(options_dict[attribute], list):
                options_dict[attribute] = [options_dict[attribute]]

        return options_dict

    @staticmethod
    def build_where_from_id(criteria: str, entity: EntityType) -> FindOneOptions | UpdateOptions:
        """Generates a FindOneOptions or UpdateOptions dictionary based on the provided criteria and entity.

        Args:
            criteria (str): The criteria to filter the entity. Typically, this is the primary key value.
            entity (EntityType): The entity type for which the options dictionary is being generated.

        Returns:
            FindOneOptions | UpdateOptions: A dictionary containing the 'where' clause for filtering the entity.

        """
        return {"where": [inspect(entity).primary_key[0] == criteria]}

    def build_options(self, pagination: Pagination) -> FindOneOptions | FindManyOptions:
        find_options = {
            "skip": pagination.skip,
            "take": pagination.take,
            "where": [],
            "order_by": [],
            "select": [],
            "relations": [],
        }

        def _make_clause(item: PaginationSearch) -> BinaryExpression:
            field = getattr(self.entity, item.get("field"), item.get("field"))
            value = item.get("value")
            return cast_if(field, String).ilike(f"%{value}%")

        search = getattr(pagination, "search", [])
        where = find_options.get("where", [])
        for param in search:
            items = param if isinstance(param, list) else [param]
            clauses = [_make_clause(it) for it in items]

            if not clauses:
                continue

            where.append(or_(*clauses) if len(clauses) > 1 else clauses[0])

        sort = getattr(pagination, "sort", [])
        order_by = find_options.get("order_by", [])
        for param in sort:
            field = getattr(self.entity, param.get("field"), param.get("field"))
            order_by.append(asc(field) if param.get("by") == "ASC" else desc(field))

        entity_relationships = inspect(self.entity).relationships
        relations = find_options.get("relations", [])
        select_options = find_options.get("select", [])
        for field in getattr(pagination, "columns", []):
            if field in entity_relationships:
                relations.append(field)
            else:
                select_options.append(getattr(self.entity, field, field))

        return find_options



import inspect
from collections.abc import Callable
from functools import singledispatchmethod, wraps
from typing import Any, TypeVar, Union, get_args, get_origin

from fastgear.common.database.sqlalchemy.session import AllSessionType, db_session

ClassType = TypeVar("ClassType")


def inject_db_parameter_decorator(cls: type[ClassType]) -> type[ClassType]:
    """Class decorator that modifies methods of the given class to automatically inject a default parameter value
    if it is not already present in the method's arguments. It applies another decorator,
    `inject_db_parameter_if_missing`, to each method.

    Args:
        cls (Type[ClassType]): The class to be decorated.

    Returns:
        Type[ClassType]: The decorated class with methods that automatically inject a default parameter value.

    """

    def _decorate_attr(attr_value: Any) -> Any:
        # staticmethod/classmethod
        if isinstance(attr_value, staticmethod | classmethod):
            fn = attr_value.__func__
            wrapped = inject_db_parameter_if_missing(fn)
            return type(attr_value)(wrapped)

        # singledispatchmethod
        if isinstance(attr_value, singledispatchmethod):
            dispatcher = attr_value.dispatcher
            # Copy and wrap each registered implementation
            for typ, fn in list(dispatcher.registry.items()):
                wrapped = inject_db_parameter_if_missing(fn)
                dispatcher.register(typ, wrapped)
            return singledispatchmethod(dispatcher)

        # Normal method or function
        if inspect.isfunction(attr_value):
            return inject_db_parameter_if_missing(attr_value)

        # Leave other attributes unchanged
        return attr_value

    for name, value in list(cls.__dict__.items()):
        # Ignore special and private attributes
        if name.startswith("__") and name.endswith("__"):
            continue
        setattr(cls, name, _decorate_attr(value))

    return cls


def inject_db_parameter_if_missing(func: Callable[..., Any]) -> Callable[..., Any]:
    """Decorator that injects a Session or AsyncSession instance into the function arguments
    if it is missing and required by the function's signature.
    """
    sig = inspect.signature(func)
    params = tuple(sig.parameters.values())
    is_coro = inspect.iscoroutinefunction(func)

    def _is_valid_session_annotation(annot: Any) -> bool:
        origin = get_origin(annot)
        if origin in (Union, getattr(__import__("types"), "UnionType", Union)):
            return any(
                isinstance(t, type) and issubclass(t, AllSessionType) for t in get_args(annot)
            )
        return isinstance(annot, type) and issubclass(annot, AllSessionType)

    # Discover the first candidate parameter for injection
    candidate_idx: int | None = None
    candidate_name: str | None = None
    for idx, p in enumerate(params):
        if (
            p.kind in (inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD)
            or p.annotation is inspect._empty
            or not _is_valid_session_annotation(p.annotation)
        ):
            continue

        if p.default is inspect._empty or p.default is None:
            candidate_idx, candidate_name = idx, p.name
            break

    # If no candidate found, return the original function
    if candidate_name is None:
        return func

    def _needs_injection(args: tuple[Any, ...], kwargs: dict[str, Any]) -> bool:
        # Already exists in args?
        if any(isinstance(a, AllSessionType) for a in args):
            return False
        # It was passed by name?
        if candidate_name in kwargs and isinstance(kwargs[candidate_name], AllSessionType):
            return False
        # Was it passed positionally?
        # and it was passed positionally and not None, do not inject
        return not (len(args) > candidate_idx and args[candidate_idx] is not None)

    if is_coro:

        @wraps(func)
        async def awrapper(*args: Any, **kwargs: Any) -> Any:
            if _needs_injection(args, kwargs):
                kwargs[candidate_name] = db_session.get()
            return await func(*args, **kwargs)

        return awrapper

    @wraps(func)
    def swrapper(*args: Any, **kwargs: Any) -> Any:
        if _needs_injection(args, kwargs):
            kwargs[candidate_name] = db_session.get()
        return func(*args, **kwargs)

    return swrapper



from typing import TypeVar

from pydantic import BaseModel

from fastgear.common.database.sqlalchemy.base import Base

EntityType = TypeVar("EntityType", bound=Base)
ColumnsQueryType = TypeVar("ColumnsQueryType", bound=type[BaseModel])
FindAllQueryType = TypeVar("FindAllQueryType", bound=type[BaseModel])
OrderByQueryType = TypeVar("OrderByQueryType", bound=type[BaseModel])



from datetime import UTC, datetime

from loguru import logger
from pydantic import BaseModel
from sqlalchemy import (
    ColumnElement,
    ForeignKeyConstraint,
    MetaData,
    Table,
    and_,
    exists,
    select,
    true,
)
from sqlalchemy.orm import registry

from fastgear.common.database.sqlalchemy.repository_utils.statement_constructor import (
    StatementConstructor,
)
from fastgear.common.database.sqlalchemy.session import SyncSessionType
from fastgear.types.generic_types_var import EntityType
from fastgear.types.http_exceptions import NotFoundException
from fastgear.types.update_options import UpdateOptions
from fastgear.types.update_result import UpdateResult

logger.bind(name="BaseRepositoryUtils")


class BaseRepositoryUtils:
    @staticmethod
    def should_be_updated(entity: EntityType, update_schema: BaseModel) -> bool:
        """Determines if the given entity should be updated based on the provided update schema.

        Args:
            entity (EntityType): The entity to check for updates.
            update_schema (BaseModel): The schema containing the update data.

        Returns:
            bool: True if the entity should be updated, False otherwise.

        """
        return any(
            getattr(entity, key) != value
            for key, value in update_schema.model_dump(exclude_unset=True).items()
        )

    @staticmethod
    def soft_delete_cascade_from_parent(
        entity: EntityType,
        *,
        update_filter: str | UpdateOptions,
        deleted_at_column="deleted_at",
        db: SyncSessionType,
    ) -> UpdateResult:
        statement_constructor = StatementConstructor(entity)

        ts = datetime.now(UTC)
        parent_table: Table = entity.__table__
        metadata = parent_table.metadata
        registry = entity.registry

        if deleted_at_column not in parent_table.c:
            raise ValueError(
                f'Parent entity "{entity.__name__}" has no "{deleted_at_column}" column'
            )

        parent_pks = list(parent_table.primary_key.columns)
        if len(parent_pks) != 1:
            raise ValueError("Composite primary keys are not supported")

        if isinstance(update_filter, str):
            update_filter = statement_constructor.build_where_from_id(update_filter, entity)
        update_filter["where"].append(parent_table.c[deleted_at_column].is_(None))

        payload = {deleted_at_column: ts}
        cmp_params = {f"cmp_{k}": v for k, v in payload.items()}
        params = {**payload, **cmp_params}
        stmt = statement_constructor.build_update_statement(
            update_filter, payload=payload
        ).returning(entity)

        result = db.execute(stmt, params)

        objs = result.scalars().all()
        affected = len(objs)

        if not affected:
            entity_name = entity.__name__
            message = f'Could not find any entity of type "{entity_name}" that matches with the search filter'
            logger.debug(message)
            raise NotFoundException(message, [entity_name])

        response = UpdateResult(raw=objs, affected=affected, generated_maps=[])

        frontier: set[Table] = {parent_table}
        visited: set[Table] = {parent_table}
        updated_tables: list[str] = [parent_table.name]

        while frontier:
            next_frontier: set[Table] = set()

            for parent in frontier:
                edges = BaseRepositoryUtils._fk_edges_from(metadata, parent)

                for child, fk in edges:
                    if child in visited:
                        continue

                    if deleted_at_column not in child.c:
                        visited.add(child)
                        next_frontier.add(child)
                        continue

                    fk_match = BaseRepositoryUtils._build_fk_match_condition(fk)
                    exists_parent_marked = exists(
                        select(1)
                        .select_from(parent)
                        .where(parent.c[deleted_at_column].is_not(None), fk_match)
                    ).correlate(child)

                    child_cls = BaseRepositoryUtils.mapped_class_for_table(child, registry)
                    stmt = statement_constructor.build_update_statement(
                        {"where": [child.c[deleted_at_column].is_(None), exists_parent_marked]},
                        payload=payload,
                        new_entity=child_cls,
                    ).returning(child_cls)
                    result = db.execute(stmt, params)

                    objs = result.scalars().all()
                    affected = len(objs)

                    if affected > 0:
                        next_frontier.add(child)
                        updated_tables.append(child.name)
                        response["affected"] += affected
                        response["raw"].extend(objs)

                    visited.add(child)

            frontier = next_frontier

        response["generated_maps"].append(updated_tables)
        return response

    @staticmethod
    def _fk_edges_from(
        metadata: MetaData, parent: Table
    ) -> list[tuple[Table, ForeignKeyConstraint]]:
        edges = []
        for table in metadata.tables.values():
            if table is parent:
                continue

            for fk in table.foreign_key_constraints:
                if fk.referred_table is parent:
                    edges.extend([(table, fk)])
        return edges

    @staticmethod
    def _build_fk_match_condition(fk: ForeignKeyConstraint) -> ColumnElement[bool]:
        conds = []
        for elem in fk.elements:
            child_col = elem.parent
            parent_col = elem.column
            conds.append(child_col == parent_col)
        return and_(*conds) if conds else true()

    @staticmethod
    def mapped_class_for_table(table: Table, registry: registry) -> type | None:
        for mapper in registry.mappers:
            if mapper.local_table is table or mapper.persist_selectable is table:
                return mapper.class_
        return None



from fastapi import Query
from fastapi_pagination.customization import CustomizedPage, UseParamsFields
from fastapi_pagination.links import Page

custom_page_query = Query(default=1, ge=1)
custom_size_query = Query(default=10, ge=1, le=100)

Page = CustomizedPage[Page, UseParamsFields(page=custom_page_query, size=custom_size_query)]



import re


class EnumUtils:
    @staticmethod
    def camel_to_snake(name: str) -> str:
        name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
        return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()

    @staticmethod
    def get_object_name(enum_cls: type) -> str:
        return EnumUtils.camel_to_snake(enum_cls.__name__)



import typing
from math import ceil
from typing import Any

from loguru import logger
from pydantic import BaseModel, TypeAdapter, ValidationError
from pydantic.fields import FieldInfo

from fastgear.types.custom_pages import Page
from fastgear.types.generic_types_var import (
    ColumnsQueryType,
    EntityType,
    FindAllQueryType,
    OrderByQueryType,
)
from fastgear.types.http_exceptions import BadRequestException
from fastgear.types.pagination import Pagination, PaginationSearch, PaginationSort


class PaginationUtils:
    def build_pagination_options(
        self,
        page: int,
        size: int,
        search: list[str] | None,
        search_all: str | None,
        sort: list[str] | None,
        columns: list[str] | None,
        columns_query: ColumnsQueryType,
        find_all_query: FindAllQueryType = None,
        order_by_query: OrderByQueryType = None,
    ) -> Pagination:
        """Build pagination options from query parameters

        Deduplicates and parses sort/search parameters, producing a Pagination
        mapping with skip, take, and normalized sort/search entries. When
        query schemas are provided, validates sort fields/directions and search
        fields/values against the corresponding Pydantic models.

        Args:
            page (int): 1-based page number used to populate the `skip` field.
            size (int): Page size used to populate the `take` field.
            search (list[str] | None): List of filters in the form "field:value". Duplicates are removed.
            search_all (str | None): Global search string applied to all searchable fields.
            sort (list[str] | None): List of sort directives in the form "field:ASC|DESC". Duplicates are removed.
            columns (list[str] | None): List of selected columns. Duplicates are removed.
            columns_query (ColumnsQueryType): Pydantic model type used to validate selectable columns.
            find_all_query (FindAllQueryType, optional): Pydantic model type used to validate search fields and values.
            order_by_query (OrderByQueryType, optional): Pydantic model type used to validate sortable fields and directions.

        Returns:
            Pagination: A pagination options mapping with keys `skip`, `take`,
            and optional `sort`/`search` lists ready for downstream processing.

        Raises:
            BadRequestException: If sort or search filters are invalid given the
            provided query schemas.
        """
        paging_options = {"skip": page, "take": size, "sort": [], "search": [], "columns": []}

        if sort:
            sort = list(set(sort))
            paging_options["sort"].extend(self._create_pagination_sort(sort))
            self._check_and_raise_for_invalid_sort_filters(paging_options["sort"], order_by_query)

        if search:
            search = list(set(search))
            paging_options["search"].extend(self._create_pagination_search(search))
            self._check_and_raise_for_invalid_search_filters(
                paging_options["search"], find_all_query
            )

        if search_all:
            paging_options["search"].extend(
                [
                    self._create_pagination_search(
                        [f"{column}:{search_all}" for column in find_all_query.model_fields]
                    )
                ]
            )

        columns = list(set(columns)) if columns else []
        paging_options["columns"] = self.select_columns(columns, columns_query)

        return Pagination(**paging_options)

    @staticmethod
    def select_columns(columns: list[str], columns_query: ColumnsQueryType) -> list[str]:
        if PaginationUtils.is_valid_column_selection(columns, columns_query):
            return PaginationUtils.merge_with_required_columns(columns, columns_query)

        message = f"Invalid columns: {columns}"
        logger.info(message)
        raise BadRequestException(message)

    @staticmethod
    def _create_pagination_sort(sort_params: list[str]) -> list[PaginationSort]:
        return [
            PaginationSort(field=field, by=by)
            for field, by in (param.split(":", 1) for param in sort_params)
        ]

    @staticmethod
    def _create_pagination_search(search_params: list[str]) -> list[PaginationSearch]:
        return [
            PaginationSearch(field=field, value=value)
            for field, value in (param.split(":", 1) for param in search_params)
        ]

    @staticmethod
    def _check_and_raise_for_invalid_sort_filters(
        pagination_sorts: list[PaginationSort], order_by_query: OrderByQueryType = None
    ) -> None:
        if order_by_query and not PaginationUtils._is_valid_sort_params(
            pagination_sorts, order_by_query
        ):
            message = f"Invalid sort filters: {pagination_sorts}"
            logger.info(message)
            raise BadRequestException(message)

    @staticmethod
    def _check_and_raise_for_invalid_search_filters(
        pagination_search: list[PaginationSearch], find_all_query: FindAllQueryType = None
    ) -> None:
        if find_all_query and not PaginationUtils._is_valid_search_params(
            pagination_search, find_all_query
        ):
            raise BadRequestException("Invalid search filters")

    @staticmethod
    def _is_valid_sort_params(
        sort: list[PaginationSort], order_by_query_schema: OrderByQueryType
    ) -> bool:
        query_schema_fields = order_by_query_schema.model_fields

        is_valid_field = all(sort_param["field"] in query_schema_fields for sort_param in sort)
        is_valid_direction = all(sort_param["by"] in ["ASC", "DESC"] for sort_param in sort)

        return is_valid_field and is_valid_direction

    @staticmethod
    def _is_valid_search_params(
        search: list[PaginationSearch], find_all_query: FindAllQueryType
    ) -> bool:
        query_dto_fields = find_all_query.model_fields

        if not PaginationUtils.validate_required_search_filter(search, query_dto_fields):
            return False

        try:
            search_params = PaginationUtils.aggregate_values_by_field(search, find_all_query)
        except KeyError as e:
            logger.info(f"Invalid search filter: {e}")
            raise BadRequestException(f"Invalid search filters: {e}")

        for search_param in search_params:
            if search_param["field"] not in query_dto_fields:
                return False

            PaginationUtils.assert_search_param_convertible(find_all_query, search_param)

        return True

    @staticmethod
    def validate_required_search_filter(
        search: list[PaginationSearch], query_dto_fields: dict[str, FieldInfo]
    ) -> bool:
        search_fields = [search_param["field"] for search_param in search]
        for field, field_info in query_dto_fields.items():
            if field_info.is_required() and field not in search_fields:
                return False

        return True

    @staticmethod
    def is_valid_column_selection(columns: list[str], columns_query_dto: ColumnsQueryType) -> bool:
        query_dto_fields = columns_query_dto.model_fields

        return all(column in query_dto_fields for column in columns)

    @staticmethod
    def merge_with_required_columns(
        columns: list[str], columns_query_dto: ColumnsQueryType
    ) -> list[str]:
        query_dto_fields = columns_query_dto.model_fields

        for field, field_info in query_dto_fields.items():
            if field_info.is_required() and field not in columns:
                columns.append(field)

        return columns

    @staticmethod
    def to_page_response(
        items: list[EntityType | BaseModel], total: int, offset: int, size: int
    ) -> Page[EntityType | BaseModel]:
        """
        Construct a Page value object containing the items for the current page
        together with pagination metadata derived from the supplied parameters.

        Args:
            items (list[EntityType | BaseModel]): Items belonging to the current page.
            total (int): Total number of items available across all pages.
            offset (int): Number of items skipped (offset). This method treats `skip`
                as an offset (0-based count of items to skip).
            size (int): Number of items per page.

        Returns:
            Page[EntityType | BaseModel]: A Page object containing the items and
            pagination metadata.

        Notes:
            - This function does not perform validation of arguments (e.g. negative
              values or zero page_size). Callers should validate inputs before use.
        """
        current_page = offset // size + 1

        return Page(
            items=items, page=current_page, size=size, total=total, pages=ceil(total / size)
        )

    @staticmethod
    def assert_no_blocked_attributes(
        block_attributes: list[str],
        search: list | None,
        sort: list | None,
        columns: list | None,
        search_all: str | None,
    ) -> None:
        """Assert that blocked pagination attributes are not present

        Checks whether any of the provided pagination attributes are present and,
        if so, logs the blocked attributes and raises BadRequestException.

        Args:
            block_attributes (list[str]): Attributes that are blocked for the route.
            search (list | None): Search filters provided by the request.
            sort (list | None): Sort directives provided by the request.
            columns (list | None): Selected columns provided by the request.
            search_all (str | None): Global search string provided by the request.

        Returns:
            None: This function only raises on violation.

        Raises:
            BadRequestException: If any blocked attribute is present in the request.
        """
        attributes_map = {
            "search": search,
            "sort": sort,
            "columns": columns,
            "search_all": search_all,
        }
        blocked = [attr for attr in block_attributes if attributes_map.get(attr) is not None]
        if not blocked:
            return

        logger.info(f"Invalid block attribute(s): {blocked}")
        raise BadRequestException(
            f"The attribute(s) {blocked} are blocked in this route and cannot be used.", loc=blocked
        )

    @staticmethod
    def assert_search_param_convertible(
        find_all_query: FindAllQueryType, search_param: PaginationSearch
    ) -> bool:
        """Validate that a search parameter value is convertible to the query type

        Attempt to validate the single-field mapping {field: value} against the provided
        Pydantic find_all_query using TypeAdapter.validate_python. On success returns
        True; on conversion failure raises BadRequestException.

        Args:
            find_all_query (FindAllQueryType): Pydantic model class describing expected field types.
            search_param (PaginationSearch): Mapping with field (str) and value (str) to validate.

        Returns:
            bool: True if the value can be converted to the expected type.

        Raises:
            BadRequestException: If the value is invalid or cannot be converted.
        """
        try:
            TypeAdapter(find_all_query).validate_python(
                {search_param["field"]: search_param["value"]}
            )
            return True
        except (ValueError, TypeError, ValidationError) as e:
            logger.info(f"Invalid search value: {e}")
            raise BadRequestException(f"Invalid search value: {e}")

    @staticmethod
    def aggregate_values_by_field(
        entries: list[PaginationSearch], find_all_query: FindAllQueryType
    ) -> list[PaginationSearch]:
        """Aggregates values by field from a list of pagination search entries.

        Args:
            entries (List[PaginationSearch]): A list of pagination search entries, each containing
            a field and value.
            find_all_query (FindAllQueryType): The query object that defines the expected types for the fields.

        Returns:
            List[PaginationSearch]: A list of PaginationSearch where each element contains a field and its aggregated values.
        """
        query_attr_types = typing.get_type_hints(find_all_query)
        aggregated: dict[str, str | list[str]] = {}

        for entry in entries:
            field, value = entry["field"], entry["value"]
            field_is_list = PaginationUtils._is_list_type_hint(query_attr_types.get(field))

            if field in aggregated:
                if isinstance(aggregated[field], list):
                    aggregated[field].append(value)
                else:
                    aggregated[field] = [aggregated[field], value]
            else:
                aggregated[field] = [value] if field_is_list else value

        return [{"field": f, "value": v} for f, v in aggregated.items()]

    @staticmethod
    def _is_list_type_hint(field_type: Any) -> bool:
        """Return whether the given type hint represents a list.

        Args:
            field_type (Any): A type hint to inspect (for example from typing.get_type_hints).

        Returns:
            bool: True if the origin of the type hint is `list`, otherwise False.

        Examples:
            >>> PaginationUtils._is_list_type_hint(list[int])
            True
        """
        return typing.get_origin(field_type) is list



from .base_schema import BaseSchema
from .custom_base_model_schema import CustomBaseModel
from .exception_response_schema import DetailResponseSchema, ExceptionResponseSchema

__all__ = ["BaseSchema", "CustomBaseModel", "DetailResponseSchema", "ExceptionResponseSchema"]



import inspect
from collections.abc import Callable
from functools import wraps
from typing import TypeVar

from fastgear.common.database.sqlalchemy.session import (
    AsyncDatabaseSessionFactory,
    SyncDatabaseSessionFactory,
    db_session,
)

T = TypeVar("T", bound=Callable)


class DBSessionDecorator:
    def __init__(
        self, session_factory: SyncDatabaseSessionFactory | AsyncDatabaseSessionFactory
    ) -> None:
        self.session_factory = session_factory

    def __call__(self, func: T) -> T:
        is_coroutine = inspect.iscoroutinefunction(func)

        @wraps(func)
        async def async_wrapper(*args, **kwargs):
            async with self.session_factory.get_session() as session:
                db_session.set(session)
                try:
                    return await func(*args, **kwargs)
                finally:
                    db_session.set(None)

        @wraps(func)
        def sync_wrapper(*args, **kwargs):
            with self.session_factory.get_session() as session:
                db_session.set(session)
                try:
                    return func(*args, **kwargs)
                finally:
                    db_session.set(None)

        return async_wrapper if is_coroutine else sync_wrapper



from datetime import datetime
from typing import Any
from uuid import UUID

from pydantic import BaseModel
from sqlalchemy.exc import InvalidRequestError

from fastgear.common.schema.custom_base_model_schema import CustomBaseModel


class BaseSchema(CustomBaseModel):
    """Base schema class that includes common fields and a method to validate and exclude unloaded fields.

    Attributes:
        id (UUID): Unique identifier for the schema.
        created_at (datetime | None): Timestamp when the record was created.
        updated_at (datetime | None): Timestamp when the record was last updated.

    """

    id: UUID
    created_at: datetime | None = None
    updated_at: datetime | None = None

    @classmethod
    def _extract_fields_to_dict(cls: type[BaseModel], obj: Any) -> dict[str, Any]:
        if isinstance(obj, dict):
            return {key: obj[key] for key in cls.model_fields if key in obj}

        result: dict[str, Any] = {}
        for field_name in cls.model_fields:
            try:
                result[field_name] = getattr(obj, field_name)
            except InvalidRequestError:
                continue
        return result

    @classmethod
    def model_validate_exclude_unloaded(cls: type[BaseModel], obj: Any) -> BaseModel:
        return cls(**cls._extract_fields_to_dict(obj))

    @classmethod
    def to_dict_exclude_unloaded(cls: type[BaseModel], obj: Any) -> dict[str, Any]:
        return cls._extract_fields_to_dict(obj)



from dataclasses import dataclass
from datetime import datetime

from sqlalchemy import DateTime, event
from sqlalchemy.orm import Mapped, ORMExecuteState, Session, with_loader_criteria
from sqlalchemy.testing.schema import mapped_column


@dataclass
class SoftDeleteMixin:
    """Mixin that provides a soft-delete timestamp column.

    Attributes:
        deleted_at (Mapped[datetime] | None): UTC timestamp (timezone-aware)
            indicating when the row was soft-deleted. A value of ``None``
            means the row is active.

    Usage:
        - Inherit from this mixin on ORM mapped classes to add a
          soft-delete column.
        - The module-level session listener will automatically filter
          out rows where ``deleted_at`` is non-null from SELECT queries.

    Example:
        class User(SoftDeleteMixin, Base):
            __tablename__ = "user"
            id: Mapped[int] = mapped_column(primary_key=True)

    """

    deleted_at: Mapped[datetime | None] = mapped_column(DateTime(timezone=True), nullable=True)


@event.listens_for(Session, "do_orm_execute", propagate=True)
def _add_filtering_criteria(execute_state: ORMExecuteState) -> None:
    """Session-level event that applies soft-delete filtering to SELECTs.

    This listener inspects ORM execution state and, for SELECT statements
    (unless explicitly disabled), applies a loader criterion so that any
    mapped class that inherits ``SoftDeleteMixin`` will only load rows
    where ``deleted_at`` is ``None``.

    Args:
        execute_state (ORMExecuteState): The ORM execution state provided
            by SQLAlchemy when emitting the "do_orm_execute" event.

    Returns:
        None

    Notes:
        To bypass the automatic soft-delete filter for a particular
        execution, use:

            session.execute(stmt, execution_options={"skip_filter": True})

    """
    with_deleted = execute_state.execution_options.get("with_deleted", False)
    if execute_state.is_select and not with_deleted:
        execute_state.statement = execute_state.statement.options(
            with_loader_criteria(
                SoftDeleteMixin, lambda cls: cls.deleted_at.is_(None), include_aliases=True
            )
        )



from dataclasses import dataclass

from typing_extensions import TypedDict


class PaginationSearch(TypedDict):
    field: str
    value: str


class PaginationSort(TypedDict):
    field: str
    by: str


@dataclass(slots=True, frozen=True)
class Pagination:
    skip: int
    take: int
    sort: list[PaginationSort]
    search: list[PaginationSearch | list[PaginationSearch]]
    columns: list[str]

    def __post_init__(self):
        # Convert page number to zero-based index for skip
        new_skip = (self.skip - 1) * self.take
        object.__setattr__(self, "skip", new_skip)



from typing import Annotated, Literal

from fastapi import Query
from pydantic import constr

from fastgear.constants import regex
from fastgear.decorators.simple_pagination_decorator import SimplePaginationOptions
from fastgear.types.custom_pages import custom_page_query, custom_size_query
from fastgear.types.generic_types_var import ColumnsQueryType, FindAllQueryType, OrderByQueryType
from fastgear.types.pagination import Pagination
from fastgear.utils import PaginationUtils

SearchString = constr(pattern=f"^{regex.ANY_CHAR}:{regex.ANY_CHAR}$")
SortString = constr(pattern=f"^{regex.ANY_CHAR}:{regex.ORDER_BY_QUERY}")
ColumnsString = constr(pattern=f"^{regex.ANY_CHAR}$")


class PaginationWithSearchOptions(SimplePaginationOptions):
    def __init__(
        self,
        columns_query: ColumnsQueryType,
        find_all_query: FindAllQueryType = None,
        order_by_query: OrderByQueryType = None,
        block_attributes: list[Literal["search", "sort", "columns", "search_all"]] = None,
    ) -> None:
        """Initializes the PaginationWithSearchOptions class.

        This constructor sets up the pagination options with search capabilities for a given entity.
        It initializes the entity, columns query, find all query, order by query, and block attributes.

        Args:
            columns_query (ColumnsQueryType): The query for selecting specific columns.
            find_all_query (FindAllQueryType, optional): The query for finding all records. Defaults to None.
            order_by_query (OrderByQueryType, optional): The query for ordering the records. Defaults to None.
            block_attributes (List[Literal["search", "sort", "columns", "search_all"]], optional):
                A list of attributes to block from being used in the pagination options. Defaults to an empty list.

        """
        if block_attributes is None:
            block_attributes = []
        self.columns_query = columns_query
        self.find_all_query = find_all_query
        self.order_by_query = order_by_query
        self.block_attributes = block_attributes

        self.pagination_utils = PaginationUtils()

    def __call__(
        self,
        page: int = custom_page_query,
        size: int = custom_size_query,
        search: Annotated[list[SearchString] | None, Query(examples=["field:value"])] = None,
        sort: Annotated[list[SortString] | None, Query(examples=["field:by"])] = None,
        columns: Annotated[list[ColumnsString] | None, Query(examples=["field"])] = None,
        search_all: Annotated[
            str | None, Query(pattern=f"^{regex.ANY_CHAR}$", examples=["value"])
        ] = None,
    ) -> Pagination:
        """Generates pagination and search options.

        This method is called to create a `FindManyOptions` object that includes pagination, sorting,
        column selection, and search parameters based on the provided arguments.

        Args:
            page (int, optional): The page number for pagination. Defaults to `custom_page_query`.
            size (int, optional): The number of items per page. Defaults to `custom_size_query`.
            search (Annotated[List[SearchString] | None, Query], optional): A list of search
                strings in the format "field:value". Defaults to None.
            sort (Annotated[List[SortString] | None, Query], optional): A list of sort strings in the format
                "field:by". Defaults to None.
            columns (Annotated[List[ColumnsString] | None, Query], optional): A list of columns to
                include in the result. Defaults to None.
            search_all (Annotated[str | None, Query], optional): A global search string to apply
                to all fields. Defaults to None.

        Returns:
            FindManyOptions: An object containing the pagination and search options.

        """
        self.pagination_utils.assert_no_blocked_attributes(
            self.block_attributes, search, sort, columns, search_all
        )

        return self.pagination_utils.build_pagination_options(
            page,
            size,
            search,
            search_all,
            sort,
            columns,
            self.columns_query,
            self.find_all_query,
            self.order_by_query,
        )



import json
from copy import deepcopy
from datetime import UTC, datetime

from fastapi import FastAPI, Request, Response
from fastapi.exceptions import RequestValidationError
from starlette.exceptions import HTTPException
from starlette.exceptions import HTTPException as StarletteHTTPException
from starlette.status import HTTP_422_UNPROCESSABLE_CONTENT

from fastgear.common.schema import DetailResponseSchema, ExceptionResponseSchema
from fastgear.types.custom_base_exception import CustomBaseException
from fastgear.types.http_exceptions import (
    CustomHTTPExceptionType,
)
from fastgear.utils import JsonUtils


class HttpExceptionsHandler:
    def __init__(self, app: FastAPI, *, add_custom_error_response: bool = False) -> None:
        self.app = app
        self.add_exceptions_handler()
        self.custom_error_response(app) if add_custom_error_response else None

    def add_exceptions_handler(self) -> None:
        """Adds exception handlers to the FastAPI app for handling various HTTP exceptions."""

        @self.app.exception_handler(StarletteHTTPException)
        async def http_exception_handler(request: Request, exc: HTTPException) -> Response:
            """Handles Starlette HTTP exceptions and returns a formatted JSON response.

            Args:
                request (Request): The incoming request object.
                exc (HTTPException): The HTTP exception raised.

            Returns:
                Response: A JSON response with the exception details.

            """
            return Response(
                status_code=exc.status_code,
                headers={"content-type": "application/json"},
                content=json.dumps(
                    self.global_exception_error_message(
                        status_code=exc.status_code,
                        detail=DetailResponseSchema(
                            loc=[], msg=exc.detail, type="Starlette HTTP Exception"
                        ),
                        request=request,
                    ).model_dump(),
                    default=JsonUtils.json_serial,
                ),
            )

        @self.app.exception_handler(RequestValidationError)
        async def validation_exception_handler(
            request: Request, exc: RequestValidationError
        ) -> Response:
            """Handles FastAPI request validation errors and returns a formatted JSON response.

            Args:
                request (Request): The incoming request object.
                exc (RequestValidationError): The validation error exception raised.

            Returns:
                Response: A JSON response with the validation error details.

            """
            return Response(
                status_code=HTTP_422_UNPROCESSABLE_CONTENT,
                headers={"content-type": "application/json"},
                content=json.dumps(
                    self.global_exception_error_message(
                        status_code=HTTP_422_UNPROCESSABLE_CONTENT,
                        detail=[DetailResponseSchema(**detail) for detail in exc.errors()],
                        request=request,
                    ).model_dump(),
                    default=JsonUtils.json_serial,
                ),
            )

        @self.app.exception_handler(CustomBaseException)
        async def custom_exceptions_handler(
            request: Request, exc: CustomHTTPExceptionType
        ) -> Response:
            """Handles custom HTTP exceptions and returns a formatted JSON response.

            This function is a FastAPI exception handler that processes various custom HTTP exceptions
            and generates a standardized JSON response containing the exception details.

            Args:
                request (Request): The incoming request object.
                exc (CustomHTTPExceptionType): The custom HTTP exception raised.

            Returns:
                Response: A JSON response with the exception details.

            """
            detail_dict = deepcopy(exc.__dict__)
            detail_dict.pop("status_code", None)

            return Response(
                status_code=exc.status_code,
                headers={"content-type": "application/json"},
                content=json.dumps(
                    self.global_exception_error_message(
                        status_code=exc.status_code,
                        detail=DetailResponseSchema(**detail_dict),
                        request=request,
                    ).model_dump(),
                    default=JsonUtils.json_serial,
                ),
            )

    @staticmethod
    def global_exception_error_message(
        status_code: int,
        detail: DetailResponseSchema | list[DetailResponseSchema],
        request: Request,
    ) -> ExceptionResponseSchema:
        """Generates a standardized error message for exceptions.

        This function creates an `ExceptionResponseSchema` object that includes details about the exception,
        such as the status code, error details, timestamp, request path, and HTTP method.

        Args:
            status_code (int): The HTTP status code of the exception.
            detail (DetailResponseSchema | List[DetailResponseSchema]): The details of the exception. Can be a single
                `DetailResponseSchema` object or a list of such objects.
            request (Request): The incoming request object that triggered the exception.

        Returns:
            ExceptionResponseSchema: An object containing the standardized error message.

        """
        if not isinstance(detail, list):
            detail = [detail]

        return ExceptionResponseSchema(
            detail=detail,
            status_code=status_code,
            timestamp=datetime.now(UTC).astimezone(),
            path=request.url.path,
            method=request.method,
        )

    @staticmethod
    def custom_error_response(app: FastAPI) -> dict:
        """Customizes the error response schema in the FastAPI app's OpenAPI documentation.

        This function modifies the default OpenAPI schema generated by FastAPI to include
        custom error response schemas. It specifically targets the 422 Validation Error response
        and replaces it with a custom schema reference.

        Args:
            app (FastAPI): The FastAPI application instance for which the OpenAPI schema will be customized.

        Returns:
            dict: The customized OpenAPI schema. If the schema has already been generated, it returns the existing
                schema.

        """
        from fastapi.openapi.constants import REF_PREFIX
        from pydantic.json_schema import models_json_schema

        original_openapi = app.openapi

        def custom_openapi():
            if app.openapi_schema:
                return app.openapi_schema

            openapi_schema = original_openapi()

            paths = openapi_schema.get("paths", {})
            for methods in paths.values():
                for details in methods.values():
                    responses = details.get("responses", {})
                    if "422" in responses:
                        responses["422"] = {
                            "description": "Validation Error",
                            "content": {
                                "application/json": {
                                    "schema": {"$ref": f"{REF_PREFIX}ExceptionResponseSchema"}
                                }
                            },
                        }

            _, defs_schema = models_json_schema(
                [(ExceptionResponseSchema, "validation")], ref_template=f"{REF_PREFIX}{{model}}"
            )
            defs = defs_schema.get("$defs", {})

            openapi_schemas = openapi_schema.setdefault("components", {}).setdefault("schemas", {})
            openapi_schemas.update(defs)
            openapi_schemas.pop("ValidationError", None)
            openapi_schemas.pop("HTTPValidationError", None)

            app.openapi_schema = openapi_schema
            return app.openapi_schema

        app.openapi = custom_openapi
        return app.openapi_schema



import inspect
from collections.abc import Callable
from typing import (
    Any,
    TypeVar,
    cast,
    get_type_hints,
)

import pydantic
from fastapi import APIRouter, Depends
from fastapi.routing import APIRoute
from starlette.routing import Route, WebSocketRoute

PYDANTIC_VERSION = pydantic.VERSION
if PYDANTIC_VERSION[0] == "2":
    from typing_inspect import is_classvar
else:
    from pydantic.typing import is_classvar  # type: ignore[no-redef]

T = TypeVar("T")

CONTROLLER_CLASS_KEY = "__controller_class__"
INCLUDE_INIT_PARAMS_KEY = "__include_init_params__"
RETURN_TYPES_FUNC_KEY = "__return_types_func__"


def controller(router: APIRouter, *urls: str) -> Callable[[type[T]], type[T]]:
    """
    This function returns a decorator that converts the decorated into a class-based view for the provided router.

    Any methods of the decorated class that are decorated as endpoints using the router provided to this function
    will become endpoints in the router. The first positional argument to the methods (typically `self`)
    will be populated with an instance created using FastAPI's dependency-injection.

    For more detail, review the documentation at
    https://fastapi-restful.netlify.app/user-guide/class-based-views//#the-cbv-decorator
    """

    def decorator(cls: type[T]) -> type[T]:
        # Define cls as controller class exclusively when using the decorator
        return _controller(router, cls, *urls)

    return decorator


def _controller(router: APIRouter, cls: type[T], *urls: str, instance: Any = None) -> type[T]:
    """
    Replaces any methods of the provided class `cls` that are endpoints of routes in `router` with updated
    function calls that will properly inject an instance of `cls`.
    """
    _init_controller(cls, instance)
    _register_endpoints(router, cls, *urls)
    return cls


def _init_controller(cls: type[Any], instance: Any = None) -> None:
    """
    Idempotently modifies the provided `cls`, performing the following modifications:
    * The `__init__` function is updated to set any class-annotated dependencies as instance attributes
    * The `__signature__` attribute is updated to indicate to FastAPI what arguments should be passed to the initializer
    """
    if getattr(cls, CONTROLLER_CLASS_KEY, False):  # pragma: no cover
        return  # Already initialized
    old_init: Callable[..., Any] = cls.__init__
    old_signature = inspect.signature(old_init)
    old_parameters = list(old_signature.parameters.values())[1:]  # drop `self` parameter
    new_parameters = [
        x
        for x in old_parameters
        if x.kind not in (inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD)
        and not _is_manually_resolved(x)
    ]

    dependency_names: list[str] = []
    for name, hint in get_type_hints(cls).items():
        if is_classvar(hint):
            continue
        parameter_kwargs = {"default": getattr(cls, name, Ellipsis)}
        dependency_names.append(name)
        new_parameters.append(
            inspect.Parameter(
                name=name, kind=inspect.Parameter.KEYWORD_ONLY, annotation=hint, **parameter_kwargs
            )
        )
    new_signature = inspect.Signature(())
    if not instance or hasattr(cls, INCLUDE_INIT_PARAMS_KEY):
        new_signature = old_signature.replace(parameters=new_parameters)

    def new_init(self: Any, *args: Any, **kwargs: Any) -> None:
        for dep_name in dependency_names:
            dep_value = kwargs.pop(dep_name)
            setattr(self, dep_name, dep_value)
        if instance and not hasattr(cls, INCLUDE_INIT_PARAMS_KEY):
            self.__class__ = instance.__class__
            self.__dict__ = instance.__dict__
        else:
            old_init(self, *args, **kwargs)

    setattr(cls, "__signature__", new_signature)
    setattr(cls, "__init__", new_init)
    setattr(cls, CONTROLLER_CLASS_KEY, True)


def _register_endpoints(router: APIRouter, cls: type[Any], *urls: str) -> None:
    controller_router = APIRouter()

    function_members = inspect.getmembers(cls, inspect.isfunction)
    for url in urls:
        _allocate_routes_by_method_name(router, url, function_members)

    router_roles = []
    for route in router.routes:
        if not isinstance(route, APIRoute):
            raise ValueError("The provided routes should be of type APIRoute")

        route_methods: Any = route.methods
        cast("tuple[Any]", route_methods)
        router_roles.append((route.path, tuple(route_methods)))

    if len(set(router_roles)) != len(router_roles):
        raise Exception("An identical route role has been implemented more then once")

    functions_set = {func for _, func in function_members}
    controller_routes = [
        route
        for route in router.routes
        if isinstance(route, Route | WebSocketRoute) and route.endpoint in functions_set
    ]
    prefix_length = len(router.prefix)  # Until 'black' would fix an issue which causes PEP8: E203

    for route in controller_routes:
        router.routes.remove(route)
        _remove_router_tags(route, router)
        route.path = route.path[prefix_length:]
        _update_controller_route_endpoint_signature(cls, route)
        _update_route_summary(route)
        if not route.name.startswith(f"{cls.__name__}."):
            route.name = f"{cls.__name__}.{route.name}"
        controller_router.routes.append(route)

    if router.prefix:
        _temp_router = APIRouter(
            tags=router.tags,
            dependencies=router.dependencies,
            callbacks=router.callbacks,
            responses=router.responses,
            default_response_class=router.default_response_class,
        )
        _temp_router.include_router(controller_router, prefix=router.prefix)
        router.routes.extend(_temp_router.routes)
    else:
        router.include_router(controller_router)


def _is_manually_resolved(param: inspect.Parameter) -> bool:
    return (
        param.default is None
        and param.annotation is not inspect.Parameter.empty
        and inspect.isclass(param.annotation)
        and param.annotation.__module__ != "builtins"
    )


def _remove_router_tags(route: Route, router: APIRouter) -> None:
    if hasattr(route, "tags"):
        route.tags = list(set(route.tags) - set(router.tags))


def _convert_to_title_case(name: str) -> str:
    return " ".join(word.capitalize() for word in name.split("_"))


def _update_route_summary(route: Route | WebSocketRoute) -> None:
    if not isinstance(route, APIRoute):
        return

    if route.summary is None:
        route.summary = _convert_to_title_case(route.endpoint.__name__)


def _allocate_routes_by_method_name(
    router: APIRouter, url: str, function_members: list[tuple[str, Any]]
) -> None:
    existing_routes_endpoints: list[tuple[Any, str]] = [
        (route.endpoint, route.path) for route in router.routes if isinstance(route, APIRoute)
    ]
    for name, func in function_members:
        if (
            hasattr(router, name)
            and not name.startswith("__")
            and not name.endswith("__")
            and (func, url) not in existing_routes_endpoints
        ):
            response_model = None
            responses = None
            kwargs = {}
            status_code = 200
            return_types_func = getattr(func, RETURN_TYPES_FUNC_KEY, None)
            if return_types_func:
                response_model, status_code, responses, kwargs = return_types_func()

            api_resource = router.api_route(
                url,
                methods=[name.capitalize()],
                response_model=response_model,
                status_code=status_code,
                responses=responses,
                **kwargs,
            )
            api_resource(func)


def _update_controller_route_endpoint_signature(
    cls: type[Any], route: Route | WebSocketRoute
) -> None:
    """
    Fixes the endpoint signature for a controller route to ensure FastAPI performs dependency injection properly.
    """
    old_endpoint = route.endpoint
    old_signature = inspect.signature(old_endpoint)
    old_parameters: list[inspect.Parameter] = list(old_signature.parameters.values())
    old_first_parameter = old_parameters[0]
    new_first_parameter = old_first_parameter.replace(default=Depends(cls))
    new_parameters = [new_first_parameter] + [
        parameter.replace(kind=inspect.Parameter.KEYWORD_ONLY) for parameter in old_parameters[1:]
    ]

    new_signature = old_signature.replace(parameters=new_parameters)
    setattr(route.endpoint, "__signature__", new_signature)



from fastgear.types.custom_pages import custom_page_query, custom_size_query
from fastgear.types.pagination import Pagination


class SimplePaginationOptions:
    def __call__(self, page: int = custom_page_query, size: int = custom_size_query) -> Pagination:
        """Generates pagination options based on the provided page and size.

        Args:
            page (int): The page number for pagination. Defaults to custom_page_query.
            size (int): The size of each page. Defaults to custom_size_query.

        Returns:
            FindManyOptions: The formatted pagination options including skip and take values.

        """
        return Pagination(
            skip=getattr(page, "default", page),
            take=getattr(size, "default", size),
            sort=[],
            search=[],
            columns=None,
        )



# Regular expression to match an ORDER BY clause with 'ASC' or 'DESC'.
ORDER_BY_QUERY = "(ASC|DESC)$"

# Regular expression to match any alphanumeric character (one or more times).
ANY_CHAR = "[a-zA-Z0-9_]+"

# Regular expression to match logical operators 'AND' or 'OR'.
OPERATOR = "(AND|OR)$"
